from kotocore.utils.constants import DEFAULT_DOCSTRING
from kotocore.exceptions import NoSuchMethod
from kotocore.loader import ResourceJSONLoader
from kotocore.utils.mangle import to_snake_case
from kotocore.utils import six


class CollectionDetails(object):
    """
    A class that encapsulates the metadata about a given ``Collection``.

    Usually hangs off a ``Collection`` as ``Collection._details``.
    """
    service_name = ''
    collection_name = ''
    session = None

    def __init__(self, session, service_name, collection_name, loader=None):
        """
        Creates a ``CollectionDetails`` instance.

        :param session: The configured ``Session`` object to refer to.
        :type session: <class kotocore.session.Session> instance

        :param service_name: The service a given ``Collection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: The name of the ``Collection``. Ex.
            ``Queue``, ``Notification``, ``Table``, etc.
        :type collection_name: string

        :param loader: (Optional) An instance of a ``ResourceJSONLoader`` class.
            This can be swapped with a different instance or with a completely
            different class with the same interface.
        :type loader: <class kotocore.loader.ResourceJSONLoader> instance
        """
        super(CollectionDetails, self).__init__()
        self.session = session
        self.service_name = service_name
        self.collection_name = collection_name
        self.loader = loader
        self._api_version = None
        self._loaded_data = None

    def __str__(self):
        return u'<{0}: {1} - {2}>'.format(
            self.__class__.__name__,
            self.service_name,
            self.collection_name,
            self.api_version
        )

    # Kinda ugly (method within a class definition, but not static/classmethod)
    # but depends on internal state. Grump.
    def requires_loaded(func):
        """
        A decorator to ensure the resource data is loaded.
        """
        def _wrapper(self, *args, **kwargs):
            # If we don't have data, go load it.
            if self._loaded_data is None:
                self._loaded_data = self.loader.load(self.service_name)

            return func(self, *args, **kwargs)

        return _wrapper

    @property
    @requires_loaded
    def service_data(self):
        """
        Returns all introspected service data. This will include things like
        other resources/collections that are part of the service. Typically,
        using ``.collection_data`` is much more useful/relevant.

        If the data has been previously accessed, a memoized version of the
        data is returned.

        :returns: A dict of introspected service data
        :rtype: dict
        """
        return self._loaded_data

    @property
    @requires_loaded
    def collection_data(self):
        """
        Returns all introspected collection data.

        If the data has been previously accessed, a memoized version of the
        data is returned.

        :returns: A dict of introspected collection data
        :rtype: dict
        """
        return self._loaded_data['collections'][self.collection_name]

    @property
    @requires_loaded
    def api_version(self):
        """
        Returns the API version introspected from the collection data.
        This is useful in preventing mismatching API versions between the
        client code & service.

        If the data has been previously accessed, a memoized version of the
        API version is returned.

        :returns: The service's version
        :rtype: string
        """
        self._api_version = self._loaded_data.get('api_version', '')
        return self._api_version

    @property
    @requires_loaded
    def resource(self):
        """
        Returns the ``Resource`` to which this collection should be mapped.

        If the data has been previously accessed, a memoized version of the
        resource name is returned.

        :returns: The resource name
        :rtype: string
        """
        return self.collection_data.get('resource', None)

    @property
    @requires_loaded
    def identifiers(self):
        """
        Returns the identifiers.

        If the data has been previously accessed, a memoized version of the
        variable name is returned.

        :returns: The identifiers
        :rtype: list
        """
        # Unlike, ``ResourceDetails``, ``identifiers`` is **optionally**
        # present on ``Collections``.
        return self.collection_data.get('identifiers', [])

    @requires_loaded
    def result_key_for(self, op_name):
        """
        Checks for the presence of a ``result_key``, which defines what data
        should make up an instance.

        Returns ``None`` if there is no ``result_key``.

        :param op_name: The operation name to look for the ``result_key`` in.
        :type op_name: string

        :returns: The expected key to look for data within
        :rtype: string or None
        """
        ops = self.collection_data.get('operations', {})
        op = ops.get(op_name, {})
        key = op.get('result_key', None)
        return key


class Collection(six.Iterator):
    """
    A common base class for all the ``Collection`` objects.
    """
    _res_class = None

    def __init__(self, connection=None, **kwargs):
        """
        Creates a new ``Collection`` instance.

        :param connection: (Optional) Specifies what connection to use.
            By default, this is a matching ``Connection`` subclass provided
            by the ``session`` (i.e. within S3, ``BucketCollection`` would get a
            ``S3Connection`` from the session).
        :type connection: <class kotocore.connection.Connection> **SUBCLASS**

        :param **kwargs: (Optional) Reserved for future use.
        :type **kwargs: dict
        """
        self._data = {}
        self._connection = connection
        self._active_iter = None
        self._active_offset = 0

        for key, value in kwargs.items():
            self._data[key] = value

        if self._connection is None:
            self._connection = self._details.session.connect_to(
                self._details.service_name
            )

        # Now that we have a connection, we can update docstrings.
        self._update_docstrings()

    def __str__(self):
        return "{0}: {1} in {2}".format(
            self.__class__.__name__,
            self._details.service_name,
            self._connection.region_name
        )

    def __getattr__(self, name):
        """
        Attempts to return instance data for a given name if available.

        :param name: The instance data's name
        :type name: string
        """
        if name in self._data:
            return self._data[name]

        raise AttributeError("No such attribute '{0}'".format(name))

    def __iter__(self):
        self._active_iter = self.each()
        self._active_offset = 0
        return iter(self)

    def __next__(self):
        res = self._active_iter[self._active_offset]
        self._active_offset += 1
        return res

    @classmethod
    def change_resource(cls, resource_class):
        """
        Updates the default ``Resource`` class created when the ``Collection``
        is returning instances.

        Default behavior (without calling this method) is that the class
        will return whatever the ``session`` can provide.

        :param resource_class: The new ``Resource`` class to use during
            construction.
        :type resource_class: class
        """
        cls._res_class = resource_class

    def _update_docstrings(self):
        """
        Runs through the operation methods & updates their docstrings if
        necessary.

        If the method has the default placeholder docstring, this will replace
        it with the docstring from the underlying connection.
        """
        ops = self._details.collection_data['operations']

        for method_name in ops.keys():
            meth = getattr(self.__class__, method_name, None)

            if not meth:
                continue

            if meth.__doc__ != DEFAULT_DOCSTRING:
                # It already has a custom docstring. Leave it alone.
                continue

            # Needs updating. So there's at least *something* vaguely useful
            # there, use the docstring from the underlying ``Connection``
            # method.
            # FIXME: We need to figure out a way to make this more useful, if
            #        possible.
            api_name = ops[method_name]['api_name']
            conn_meth = getattr(self._connection, to_snake_case(api_name))

            # We need to do detection here, because Py2 treats ``.__doc__``
            # as a special read-only attribute. :/
            if six.PY3:
                meth.__doc__ = conn_meth.__doc__
            else:
                meth.__func__.__doc__ = conn_meth.__doc__

    def get_identifiers(self):
        """
        Returns the identifier(s) (if present) from the instance data.

        The identifier name(s) is/are determined from the ``ResourceDetails``
        instance hanging off the class itself.

        :returns: All the identifier information
        :rtype: dict
        """
        data = {}

        for id_info in self._details.identifiers:
            var_name = id_info['var_name']
            data[var_name] = self._data.get(var_name)

        return data

    def set_identifiers(self, data):
        """
        Sets the identifier(s) within the instance data.

        The identifier name(s) is/are determined from the ``ResourceDetails``
        instance hanging off the class itself.

        :param data: The value(s) to be set.
        :param data: dict
        """
        for id_info in self._details.identifiers:
            var_name = id_info['var_name']
            self._data[var_name] = data.get(var_name)

    def full_update_params(self, conn_method_name, params):
        """
        When a API method on the collection is called, this goes through the
        params & run a series of hooks to allow for updating those parameters.

        Typically, this method is **NOT** call by the user. However, the user
        may wish to define other methods (i.e. ``update_params`` to work with
        multiple parameters at once or ``update_params_METHOD_NAME`` to
        manipulate a single parameter) on their class, which this method
        will call.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param params: A dictionary of all the key/value pairs passed to the
            method. This dictionary is transformed by this call into the final
            params to be passed to the underlying connection.
        :type params: dict
        """
        # We'll check for custom methods to do addition, specific work.
        custom_method_name = 'update_params_{0}'.format(conn_method_name)
        custom_method = getattr(self, custom_method_name, None)

        if custom_method:
            # Let the specific method further process the data.
            params = custom_method(params)

        # Now that all the method-specific data is there, apply any further
        # service-wide changes here.
        params = self.update_params(conn_method_name, params)
        return params

    def update_params(self, conn_method_name, params):
        """
        A hook to allow manipulation of multiple parameters at once.

        By default, this just ensures the identifier data in in the parameters,
        so that the user doesn't have to provide it.

        You can override/extend this method (typically on your subclass)
        to do additional checks, pre-populate values or remove unwanted data.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param params: A dictionary of all the key/value pairs passed to the
            method. This dictionary is transformed by this call into the final
            params to be passed to the underlying connection.
        :type params: dict
        """
        # By default, this just sets the identifier info.
        # We use ``var_name`` instead of ``api_name``. Because botocore.
        params.update(self.get_identifiers())
        return params

    def full_post_process(self, conn_method_name, result):
        """
        When a response from an API method call is received, this goes through
        the returned data & run a series of hooks to allow for handling that
        data.

        Typically, this method is **NOT** call by the user. However, the user
        may wish to define other methods (i.e. ``post_process`` to work with
        all the data at once or ``post_process_METHOD_NAME`` to
        handle a single piece of data) on their class, which this method
        will call.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param result: A dictionary of all the key/value pairs passed back
            from the API (server-side). This dictionary is transformed by this
            call into the final data to be passed back to the user.
        :type result: dict
        """
        result = self.post_process(conn_method_name, result)

        # We'll check for custom methods to do addition, specific work.
        custom_method_name = 'post_process_{0}'.format(conn_method_name)
        custom_method = getattr(self, custom_method_name, None)

        if custom_method:
            # Let the specific method further process the data.
            result = custom_method(result)

        return result

    def post_process(self, conn_method_name, result):
        """
        A hook to allow manipulation of the entire returned data at once.

        By default, this goes through & (shallowly) snake-cases all of the
        keys, so that the result is friendlier to use from Python.

        You can override/extend this method (typically on your subclass)
        to do additional checks, alter the result or remove unwanted data.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param result: A dictionary of all the key/value pairs passed back
            from the API (server-side). This dictionary is transformed by this
            call into the final data to be passed back to the user.
        :type result: dict
        """
        return result

    def post_process_create(self, result):
        """
        An example of the ``post_process`` extensions, this returns an instance
        of the ``Resource`` created (rather than just a bag of data).

        :param result: The full data handed back from the API.
        :type result: dict

        :returns: A ``Resource`` subclass
        """
        # We need to possibly drill into the response & get out the data here.
        # Check for a result key.
        result_key = self._details.result_key_for('create')

        if not result_key:
            return self.build_resource(result)

        return self.build_resource(result[result_key])

    def post_process_each(self, result):
        """
        An example of the ``post_process`` extensions, this returns a set
        of instances of the ``Resource`` fetched (rather than just a bag of
        data).

        :param result: The full data handed back from the API.
        :type result: dict

        :returns: A list of ``Resource`` subclass
        """
        # We need to possibly drill into the response & get out the data here.
        # Check for a result key.
        result_key = self._details.result_key_for('each')

        if not result_key:
            return result

        return [self.build_resource(res) for res in result[result_key]]

    def build_resource(self, data):
        """
        Given some data, builds the correct/matching ``Resource`` subclass
        for the ``Collection``. Useful in things like ``create`` & ``each``.

        :param result: The data for an instance handed back from the API.
        :type result: dict

        :returns: A ``Resource`` subclass
        """
        if self._res_class is None:
            self._res_class = self._details.session.get_resource(
                self._details.service_name,
                self._details.resource
            )

        final_data = {}

        # Lightly post-process the data, to look more Pythonic.
        for key, value in data.items():
            final_data[to_snake_case(key)] = value

        return self._res_class(connection=self._connection, **final_data)


class CollectionFactory(object):
    """
    Generates the underlying ``Collection`` classes based off the
    ``ResourceJSON`` included in the SDK.

    Usage::

        >>> cf = CollectionFactory()
        >>> BucketCollection = cf.construct_for('s3', 'BucketCollection')

    """
    loader_class = ResourceJSONLoader

    def __init__(self, session=None, loader=None,
                 base_collection_class=Collection,
                 details_class=CollectionDetails):
        """
        Creates a new ``CollectionFactory`` instance.

        :param session: The ``Session`` the factory should use.
        :type session: <class kotocore.session.Session> instance

        :param loader: (Optional) An instance of a ``ResourceJSONLoader`` class.
            This can be swapped with a different instance or with a completely
            different class with the same interface.
            By default, this is ``kotocore.loader.default_loader``.
        :type loader: <class kotocore.loader.ResourceJSONLoader> instance

        :param base_collection_class: (Optional) The base class to use when
            creating the collection. By default, this is ``Collection``, but
            should you need to globally change the behavior of all collections,
            you'd simply specify this to provide your own class.
        :type base_collection_class: <class kotocore.collections.Collection>

        :param details_class: (Optional) The metadata class used to store things
            like service name & data. By default, this is ``CollectionDetails``,
            but should you need to globally change the behavior (perhaps
            modifying how the collection data is returned), you simply provide
            your own class here.
        :type details_class: <class kotocore.collections.CollectionDetails>
        """
        self.session = session
        self.loader = loader
        self.base_collection_class = base_collection_class
        self.details_class = details_class

        if self.session is None:
            # Fallback to the default.
            import kotocore
            self.session = kotocore.session

        if self.loader is None:
            import kotocore.loader
            self.loader = kotocore.loader.default_loader

    def __str__(self):
        return self.__class__.__name__

    def construct_for(self, service_name, collection_name, base_class=None):
        """
        Builds a new, specialized ``Collection`` subclass as part of a given
        service.

        This will load the ``ResourceJSON``, determine the correct
        mappings/methods & constructs a brand new class with those methods on
        it.

        :param service_name: The name of the service to construct a resource
            for. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: The name of the ``Collection``. Ex.
            ``QueueCollection``, ``NotificationCollection``,
            ``TableCollection``, etc.
        :type collection_name: string

        :returns: A new collection class for that service
        """
        details = self.details_class(
            self.session,
            service_name,
            collection_name,
            loader=self.loader
        )

        attrs = {
            '_details': details,
        }

        # Determine what we should call it.
        klass_name = self._build_class_name(collection_name)

        # Construct what the class ought to have on it.
        attrs.update(self._build_methods(details))

        if base_class is None:
            base_class = self.base_collection_class

        # Create the class.
        return type(
            klass_name,
            (base_class,),
            attrs
        )

    def _build_class_name(self, collection_name):
        return collection_name

    def _build_methods(self, details):
        attrs = {}
        ops = details.collection_data.get('operations', {}).items()

        for method_name, op_data in ops:
            attrs[method_name] = self._create_operation_method(
                method_name,
                op_data
            )

        return attrs

    def _create_operation_method(factory_self, method_name, op_data):
        # Determine the correct name for the method.
        # This is because the method names will be standardized across
        # resources, so we'll have to lean on the ``api_name`` to figure out
        # what the correct underlying method name on the ``Connection`` should
        # be.
        # Map -> map -> unmap -> remap -> map :/
        conn_method_name = to_snake_case(op_data['api_name'])

        if not six.PY3:
            method_name = str(method_name)

        def _new_method(self, **kwargs):
            params = self.full_update_params(method_name, kwargs)
            method = getattr(self._connection, conn_method_name, None)

            if not method:
                msg = "Introspected method named '{0}' not available on " + \
                      "the connection."
                raise NoSuchMethod(msg.format(conn_method_name))

            result = method(**params)
            return self.full_post_process(method_name, result)

        _new_method.__name__ = method_name
        _new_method.__doc__ = DEFAULT_DOCSTRING
        return _new_method



import re

from kotocore.utils.constants import DEFAULT_REGION
from kotocore.utils import six
from kotocore.utils.mangle import html_to_rst


class Introspection(object):
    """
    Used to introspect ``botocore`` objects (``Service``, ``Operation`` &
    ``Endpoint``) to determine what operations/parameters/etc. are available.

    Done here to encapsulate the data needed from ``botocore``, to limit
    API changes elsewhere in ``boto`` should ``botocore`` change.

    Usage::

        >>> from botocore.session import Session
        >>> session = Session()
        >>> intro = Introspection(session)
        >>> intro.introspect_service('s3')
        {
            # ...big bag of operation data...
        }

    """
    tag_re = re.compile(r'<.*?>')

    def __init__(self, session):
        """
        Creates a new ``Introspection`` instance.

        :param session: A ``botocore`` session to introspect with.
        :type session: A ``<botocore.session.Session>`` instance
        """
        super(Introspection, self).__init__()
        # TODO: For now, this is a ``botocore.session.Session``. We may want
        #       to use a ``kotocore.session.Session`` instead?
        self.session = session

    def get_service(self, service_name):
        """
        Returns a ``botocore.service.Service`` object for a given service.

        :param service_name: The desired service name
        :type service_name: string

        :returns: A <botocore.service.Service> object
        """
        return self.session.get_service(service_name)

    def get_endpoint(self, service, region_name=DEFAULT_REGION):
        """
        Returns a ``botocore.endpoint.Endpoint`` object for a given service.

        :param service: A ``Service`` object
        :type service: A ``<botocore.service.Service>`` instance

        :param region_name: (Optional) The name of the region you'd like to
            connect to. By default, this is
            ``kotocore.utils.constants.DEFAULT_REGION``.
        :type region_name: string

        :returns: A <botocore.endpoint.Endpoint> object
        """
        return service.get_endpoint(region_name=region_name)

    def get_operation(self, service, operation_name):
        """
        Returns a ``botocore.operation.Operation`` object for a given service
        & operation name.

        :param service_name: The desired service name
        :type service_name: string

        :param operation_name: The API name of the operation you'd like to
            execute. Ex. ``PutBucketCORS``
        :type operation_name: string

        :returns: A <botocore.operation.Operation> object
        """
        return service.get_operation(operation_name)

    def strip_html(self, doc):
        """
        This method removes all HTML from a docstring.

        Lighter than ``convert_docs``, this is intended for the documentation
        on **parameters**, not the overall docs themselves.

        :param doc: The initial docstring
        :type doc: string

        :returns: The stripped/cleaned docstring
        :rtype: string
        """
        if not isinstance(doc, six.string_types):
            return ''

        doc = doc.strip()
        doc = self.tag_re.sub('', doc)
        return doc

    def parse_param(self, core_param):
        """
        Returns data about a specific parameter.

        :param core_param: The ``Parameter`` to introspect
        :type core_param: A ``<botocore.parameters.Parameter>`` subclass

        :returns: A dict of the relevant information
        """
        return {
            'var_name': core_param.py_name,
            'api_name': core_param.name,
            'required': core_param.required,
            'docs': self.strip_html(core_param.documentation),
            'type': core_param.type,
        }

    def parse_params(self, core_params):
        """
        Goes through a set of parameters, extracting information about each.

        :param core_params: The collection of parameters
        :type core_params: A collection of ``<botocore.parameters.Parameter>``
            subclasses

        :returns: A list of dictionaries
        """
        params = []

        for core_param in core_params:
            params.append(self.parse_param(core_param))

        return params

    def convert_docs(self, html):
        """
        Converts the service's HTML docs to reStructured Text.

        :param html: The HTML to convert
        :type html: string

        :returns: The converted text
        :rtype: string
        """
        return html_to_rst(html)

    def introspect_operation(self, operation):
        """
        Introspects an entire operation, returning::

        * the method name (to expose to the user)
        * the API name (used server-side)
        * docs
        * introspected information about the parameters
        * information about the output

        :param operation: The operation to introspect
        :type operation: A <botocore.operation.Operation> object

        :returns: A dict of information
        """
        return {
            'method_name': operation.py_name,
            'api_name': operation.name,
            'docs': self.convert_docs(operation.documentation),
            'params': self.parse_params(operation.params),
            'output': operation.output,
        }

    def introspect_service(self, service_name):
        """
        Introspects all the operations (& related information) about a service.

        :param service_name: The desired service name
        :type service_name: string

        :returns: A dict of all operation names & information
        """
        data = {}
        service = self.get_service(service_name)

        for operation in service.operations:
            # These are ``Operation`` objects, not operation strings.
            op_data = self.introspect_operation(operation)
            data[op_data['method_name']] = op_data

        return data



from bcdoc.restdoc import ReSTDocument

from botocore import xform_name


def to_snake_case(camel_case_name):
    """
    Converts CamelCaseNames to snake_cased_names.

    :param camel_case_name: The name you'd like to convert from.
    :type camel_case_name: string

    :returns: A converted string
    :rtype: string
    """
    return xform_name(camel_case_name)


def to_camel_case(snake_case_name):
    """
    Converts snake_cased_names to CamelCaseNames.

    :param snake_case_name: The name you'd like to convert from.
    :type snake_case_name: string

    :returns: A converted string
    :rtype: string
    """
    bits = snake_case_name.split('_')
    return ''.join([bit.capitalize() for bit in bits])


def html_to_rst(html):
    """
    Converts the service HTML docs to reStructured Text, for use in docstrings.

    :param html: The raw HTML to convert
    :type html: string

    :returns: A reStructured Text formatted version of the text
    :rtype: string
    """
    doc = ReSTDocument()
    doc.include_doc_string(html)
    raw_doc = doc.getvalue()
    return raw_doc.decode('utf-8')



from kotocore.utils.constants import DEFAULT_REGION
from kotocore.utils.constants import NOTHING_PROVIDED
from kotocore.exceptions import ServerError
from kotocore.introspection import Introspection
from kotocore.utils import six


class ConnectionDetails(object):
    """
    A class that encapsulates the metadata about a given ``Connection``.

    Usually hangs off a ``Connection`` as ``Connection._details``.
    """
    service_name = 'unknown'
    session = None

    def __init__(self, service_name, session):
        """
        Creates a ``ConnectionDetails`` instance.

        :param service_name: The service a given ``Connection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param session: The configured ``Session`` object to refer to.
        :type session: <class kotocore.session.Session> instance
        """
        super(ConnectionDetails, self).__init__()
        self.service_name = service_name
        self.session = session
        self._api_version = None
        self._loaded_service_data = None

    def __str__(self):
        return u'<{0}: {1} - {2}>'.format(
            self.__class__.__name__,
            self.service_name,
            self.api_version
        )

    @property
    def service_data(self):
        """
        Returns all introspected service data.

        If the data has been previously accessed, a memoized version of the
        data is returned.

        :returns: A dict of introspected service data
        :rtype: dict
        """
        # Lean on the cache first.
        if self._loaded_service_data is not None:
            return self._loaded_service_data

        # We don't have a cache. Build it.
        self._loaded_service_data = self._introspect_service(
            # We care about the ``botocore.session`` here, not the
            # ``kotocore.session``.
            self.session.core_session,
            self.service_name
        )
        # Clear out the API version, just in case.
        self._api_version = None
        return self._loaded_service_data

    @property
    def api_version(self):
        """
        Returns API version introspected from the service data.

        If the data has been previously accessed, a memoized version of the
        API version is returned.

        :returns: The service's version
        :rtype: string
        """
        # Lean on the cache first.
        if self._api_version is not None:
            return self._api_version

        # We don't have a cache. Build it.
        self._api_version = self._introspect_api_version(
            self.session.core_session,
            self.service_name
        )
        return self._api_version

    def _introspect_service(self, core_session, service_name):
        # Yes, we could lean on ``self.session|.service_name`` here,
        # but this makes testing/composability easier.
        intro = Introspection(core_session)
        return intro.introspect_service(service_name)

    def _introspect_api_version(self, core_session, service_name):
        intro = Introspection(core_session)
        service = intro.get_service(service_name)
        return service.api_version

    def reload_service_data(self):
        """
        Wipes out & reloads the cached service data.

        :returns: A dict of introspected service data
        :rtype: dict
        """
        self._loaded_service_data = None
        return self.service_data


class Connection(object):
    """
    A common base class for all the ``Connection`` objects.
    """
    def __init__(self, region_name=DEFAULT_REGION):
        """
        Creates a new connection instance.

        :param region_name: (Optional) The name of the region to connect to.
            By default, this is the value from
            ``kotocore.utils.constants.DEFAULT_REGION``.
        :type region_name: string
        """
        super(Connection, self).__init__()
        self.region_name = region_name

    def __str__(self):
        return u'<{0}: {0}>'.format(
            self.__class__.__name__,
            self.region_name
        )

    def _check_method_params(self, op_params, **kwargs):
        # For now, we don't type-check or anything, just check for required
        # params.
        for param in op_params:
            if param['required'] is True:
                if not param['var_name'] in kwargs:
                    err = "Missing required parameter: '{0}'".format(
                        param['var_name']
                    )
                    raise TypeError(err)

    def _build_service_params(self, op_params, **kwargs):
        # TODO: Maybe build in an extension mechanism (like
        #      ``build_<op_name>_params``)?
        service_params = {}

        for param in op_params:
            value = kwargs.get(param['var_name'], NOTHING_PROVIDED)

            if value is NOTHING_PROVIDED:
                # They didn't give us a value. We should've already checked
                # "required-ness", so just give it a pass & move on.
                continue

            # FIXME: This is weird. I was expecting this to be
            #        ``param['api_name']`` to pass to ``botocore``, but
            #        evidently it expects snake_case here?!
            service_params[param['var_name']] = value

        return service_params

    def _check_for_errors(self, results):
        result_data = results[1]

        if 'Errors' in result_data:
            errs = result_data['Errors']

            if not errs:
                # Skip it if the errors are empty.
                # For instance, S3 will send this key with nothing in it on a
                # successful call.
                return

            if isinstance(errs, (list, tuple)):
                error = errs[0]
            elif hasattr(errs, 'items'):
                error = errs
            else:
                error = {
                    'Message': errs
                }

            raise ServerError(
                code=error.get('Code', 'ConnectionError'),
                message=error.get('Message', 'No details available.'),
                full_response=result_data
            )

    def _post_process_results(self, method_name, output, results):
        # TODO: Maybe build in an extension mechanism (like
        #      ``post_process_<op_name>_results``)?
        return results[1]

    @classmethod
    def connect_to(cls, **kwargs):
        """
        Instantiates the class, passing all ``**kwargs`` along to the
        constructor.

        This is reserved for further extension.

        :returns: An instance of the ``Connection``
        :rtype: <class Connection>
        """
        return cls(**kwargs)

    def _get_operation_data(self, method_name):
        """
        Returns all the introspected operation data for a given method.
        """
        return self._details.service_data[method_name]

    def _get_operation_params(self, method_name):
        return self._get_operation_data(method_name).get('params', [])


class ConnectionFactory(object):
    """
    Builds custom ``Connection`` subclasses based on the service's operations.

    Usage::

        >>> cf = ConnectionFactory()
        >>> S3Connection = cf.construct_for('s3')

    """
    def __init__(self, session, base_connection=Connection,
                 details_class=ConnectionDetails):
        """
        Creates a new ``ConnectionFactory`` instance.

        :param session: The ``Session`` the factory should use.
        :type session: <class kotocore.session.Session> instance

        :param base_connection: (Optional) The base class to use when creating
            the connection. By default, this is ``Connection``, but should
            you need to globally change the behavior of all connections,
            you'd simply specify this to provide your own class.
        :type base_connection: <class kotocore.connection.Connection>

        :param details_class: (Optional) The metadata class used to store things
            like service name & data. By default, this is ``ConnectionDetails``,
            but should you need to globally change the behavior (perhaps
            modifying how the service data is returned), you simply provide
            your own class here.
        :type details_class: <class kotocore.connection.ConnectionDetails>
        """
        super(ConnectionFactory, self).__init__()
        self.session = session
        self.base_connection = base_connection
        self.details_class = ConnectionDetails

    def __str__(self):
        return self.__class__.__name__

    def construct_for(self, service_name):
        """
        Builds a new, specialized ``Connection`` subclass for a given service.

        This will introspect a service, determine all the API calls it has &
        constructs a brand new class with those methods on it.

        :param service_name: The name of the service to construct a connection
            for. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :returns: A new connection class for that service
        """
        # Construct a new ``ConnectionDetails`` (or similar class) for storing
        # the relevant details about the service & its operations.
        details = self.details_class(service_name, self.session)
        # Make sure the new class gets that ``ConnectionDetails`` instance as a
        # ``cls._details`` attribute.
        attrs = {
            '_details': details,
        }

        # Determine what we should call it.
        klass_name = self._build_class_name(service_name)

        # Construct what the class ought to have on it.
        attrs.update(self._build_methods(details))

        # Create the class.
        return type(
            klass_name,
            (self.base_connection,),
            attrs
        )

    def _build_class_name(self, service_name):
        return '{0}Connection'.format(service_name.capitalize())

    def _build_methods(self, details):
        attrs = {}

        for method_name, op_data in details.service_data.items():
            # First we make expand then we defense it.
            # Construct a brand-new method & assign it on the class.
            attrs[method_name] = self._create_operation_method(method_name, op_data)

        return attrs

    def _generate_docstring(self, op_data):
        docstring = op_data['docs']

        for param_data in op_data['params']:
            param_doc = ":param {0}: {1}\n".format(
                param_data['var_name'],
                param_data.get('docs', 'No documentation available')
            )
            type_doc = ":type {0}: {1}\n".format(
                param_data['var_name'],
                param_data['type']
            )

            docstring += '\n'
            docstring += param_doc
            docstring += type_doc

        docstring += '\n'
        docstring += ':returns: The response data received\n'
        docstring += ':rtype: dict\n'
        return docstring

    def _create_operation_method(factory_self, method_name, orig_op_data):
        if not six.PY3:
            method_name = str(method_name)

        def _new_method(self, **kwargs):
            # Fetch the information about the operation.
            op_data = self._get_operation_data(method_name)

            # Check the parameters.
            self._check_method_params(
                op_data['params'],
                **kwargs
            )

            # Prep the service's parameters.
            service_params = self._build_service_params(
                op_data['params'],
                **kwargs
            )

            # Actually call the service.
            service = self._details.session.get_core_service(
                self._details.service_name
            )
            endpoint = service.get_endpoint(self.region_name)
            op = service.get_operation(
                op_data['api_name']
            )
            results = op.call(endpoint, **service_params)

            # Check for error conditions.
            self._check_for_errors(results)

            # Post-process results here
            post_processed = self._post_process_results(
                method_name,
                op_data['output'],
                results
            )
            return post_processed

        # Swap the name, so it looks right.
        _new_method.__name__ = method_name
        # Assign docstring.
        _new_method.__doc__ = factory_self._generate_docstring(orig_op_data)
        # Return the newly constructed method.
        return _new_method



import importlib

from kotocore.exceptions import IncorrectImportPath


def import_class(import_path):
    """
    Imports a class dynamically from a full import path.
    """
    if not '.' in import_path:
        raise IncorrectImportPath(
            "Invalid Python-style import path provided: {0}.".format(
                import_path
            )
        )

    path_bits = import_path.split('.')
    mod_path = '.'.join(path_bits[:-1])
    klass_name = path_bits[-1]

    try:
        mod = importlib.import_module(mod_path)
    except ImportError:
        raise IncorrectImportPath(
            "Could not import module '{0}'.".format(mod_path)
        )

    try:
        klass = getattr(mod, klass_name)
    except AttributeError:
        raise IncorrectImportPath(
            "Imported module '{0}' but could not find class '{1}'.".format(
                mod_path,
                klass_name
            )
        )

    return klass



import glob
import os

from kotocore.utils.constants import DEFAULT_RESOURCE_JSON_DIR
from kotocore.exceptions import NoResourceJSONFound
from kotocore.utils import json


class ResourceJSONLoader(object):
    """
    Handles the loading of the ResourceJSON. Can be overridden to look up
    user-defined paths first, with fallbacks.

    For optimal efficiency, one loader instance should be used in as many
    places as possible & use the ``__getitem__`` (dictionary-style) lookups.
    This will populate the cache, causing subsequent lookups to be fast & one
    instance will prevent memory bloat.

    Usage::

        >>> rjl = ResourceJSONLoader()
        # Load it no matter what.
        >>> rjl.load('s3')
        {
            # ...S3's ResourceJSON as a dict...
        }
        # Get a cached version (or load it if not present).
        >>> rjl['ec2']
        {
            # ...EC2's ResourceJSON as a dict...
        }

    """
    default_data_dirs = [
        DEFAULT_RESOURCE_JSON_DIR,
    ]

    def __init__(self, data_dirs=None):
        """
        Creates a new ``ResourceJSONLoader`` instance.

        :param data_dirs: (Optional) A list of absolute paths to check for the
            ResourceJSON. By default, this is just
            ``kotocore.utils.constants.DEFAULT_RESOURCE_JSON_DIR`` as the single
            item in the list.
        :type data_dirs: list
        """
        self.data_dirs = data_dirs
        self._loaded_data = {}

        if self.data_dirs is None:
            self.data_dirs = self.default_data_dirs

    def get_available_options(self, service_name):
        """
        Fetches a collection of all JSON files for a given service.

        This checks user-created files (if present) as well as including the
        default service files.

        Example::

            >>> loader.get_available_options('s3')
            {
                '2013-11-27': [
                    '~/.boto-overrides/s3-2013-11-27.json',
                    '/path/to/kotocore/data/aws/resources/s3-2013-11-27.json',
                ],
                '2010-10-06': [
                    '/path/to/kotocore/data/aws/resources/s3-2010-10-06.json',
                ],
                '2007-09-15': [
                    '~/.boto-overrides/s3-2007-09-15.json',
                ],
            }

        :param service_name: The name of the desired service
        :type service_name: string

        :returns: A dictionary of api_version keys, with a list of filepaths
            for that version (in preferential order).
        :rtype: dict
        """
        options = {}

        for data_dir in self.data_dirs:
            # Traverse all the directories trying to find the best match.
            service_glob = "{0}-*.json".format(service_name)
            path = os.path.join(data_dir, service_glob)
            found = glob.glob(path)

            for match in found:
                # Rip apart the path to determine the API version.
                base = os.path.basename(match)
                bits = os.path.splitext(base)[0].split('-', 1)

                if len(bits) < 2:
                    continue

                api_version = bits[1]
                options.setdefault(api_version, [])
                options[api_version].append(match)

        return options

    def get_best_match(self, options, service_name, api_version=None):
        """
        Given a collection of possible service options, selects the best match.

        If no API version is provided, the path to the most recent API version
        will be returned. If an API version is provided & there is an exact
        match, the path to that version will be returned. If there is no exact
        match, an attempt will be made to find a compatible (earlier) version.

        In all cases, user-created files (if present) will be given preference
        over the default included versions.

        :param options: A dictionary of options. See
            ``.get_available_options(...)``.
        :type options: dict

        :param service_name: The name of the desired service
        :type service_name: string

        :param api_version: (Optional) The desired API version to load
        :type service_name: string

        :returns: The full path to the best matching JSON file
        """
        if not options:
            msg = "No JSON files provided. Please check your " + \
                  "configuration/install."
            raise NoResourceJSONFound(msg)

        if api_version is None:
            # Give them the very latest option.
            best_version = max(options.keys())
            return options[best_version][0], best_version

        # They've provided an api_version. Try to give them exactly what they
        # requested, falling back to the best compatible match if no exact
        # match can be found.
        if api_version in options:
            return options[api_version][0], api_version

        # Find the best compatible match. Run through in descending order.
        # When we find a version that's lexographically less than the provided
        # one, run with it.
        for key in sorted(options.keys(), reverse=True):
            if key <= api_version:
                return options[key][0], key

        raise NoResourceJSONFound(
            "No compatible JSON could be loaded for {0} ({1}).".format(
                service_name,
                api_version
            )
        )

    def load(self, service_name, api_version=None, cached=True):
        """
        Loads the desired JSON for a service. (uncached)

        This will fall back through all the ``data_dirs`` provided to the
        constructor, returning the **first** one it finds.

        :param service_name: The name of the desired service
        :type service_name: string

        :param api_version: (Optional) The desired API version to load
        :type service_name: string

        :param cached: (Optional) Whether or not the cache should be used
            when attempting to load the data. Default is ``True``.
        :type cached: boolean

        :returns: The loaded JSON as a dict
        """
        # Fetch from the cache first if it's there.
        if cached:
            if service_name in self._loaded_data:
                if api_version in self._loaded_data[service_name]:
                    return self._loaded_data[service_name][api_version]

        data = {}
        options = self.get_available_options(service_name)
        match, version = self.get_best_match(
            options,
            service_name,
            api_version=api_version
        )

        with open(match, 'r') as json_file:
            data = json.load(json_file)
            # Embed where we found it from for debugging purposes.
            data['__file__'] = match
            data['api_version'] = version

        if cached:
            self._loaded_data.setdefault(service_name, {})
            self._loaded_data[service_name][api_version] = data

        return data

    def __contains__(self, service_name):
        return service_name in self._loaded_data


# Default instance for convenience.
default_loader = ResourceJSONLoader()



import os
import sys

from kotocore import get_version


# TODO: Assert if this is egg-safe (or if that matters to us)?
KOTOCORE_ROOT = os.path.dirname(os.path.dirname(__file__))

USER_AGENT_NAME = 'kotocore'
USER_AGENT_VERSION = get_version(full=True)

DEFAULT_REGION = 'us-east-1'

DEFAULT_DOCSTRING = """
Please make an instance of this class to inspect the docstring.

No underlying connection is yet available.
"""

DEFAULT_DATA_DIR = os.path.join(KOTOCORE_ROOT, 'data', 'aws')
DEFAULT_RESOURCE_JSON_DIR = os.path.join(DEFAULT_DATA_DIR, 'resources')


class NOTHING_PROVIDED(object):
    """
    An identifier for no data provided.

    Never meant to be instantiated.
    """
    pass


class NO_NAME(object):
    """
    An identifier to indicate a method instance hasn't been given a name.

    Never meant to be instantiated.
    """
    pass


class NO_RESOURCE(object):
    """
    An identifier to indicate a method instance hasn't been attached to a
    resource.

    Never meant to be instantiated.
    """
    pass



from kotocore.utils.constants import DEFAULT_DOCSTRING
from kotocore.exceptions import NoSuchMethod, NoRelation
from kotocore.introspection import Introspection
from kotocore.loader import ResourceJSONLoader
from kotocore.utils.mangle import to_snake_case
from kotocore.utils import six


class ResourceDetails(object):
    """
    A class that encapsulates the metadata about a given ``Resource``.

    Usually hangs off a ``Resource`` as ``Resource._details``.
    """
    service_name = ''
    resource_name = ''
    session = None

    def __init__(self, session, service_name, resource_name, loader=None):
        """
        Creates a ``ResourceDetails`` instance.

        :param session: The configured ``Session`` object to refer to.
        :type session: <class kotocore.session.Session> instance

        :param service_name: The service a given ``Resource`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param resource_name: The name of the ``Resource``. Ex.
            ``Queue``, ``Notification``, ``Table``, etc.
        :type resource_name: string

        :param loader: (Optional) An instance of a ``ResourceJSONLoader`` class.
            This can be swapped with a different instance or with a completely
            different class with the same interface.
        :type loader: <class kotocore.loader.ResourceJSONLoader> instance
        """
        super(ResourceDetails, self).__init__()
        self.session = session
        self.service_name = service_name
        self.resource_name = resource_name
        self.loader = loader
        self._api_version = None
        self._loaded_data = None

    def __str__(self):
        return u'<{0}: {1} - {2}>'.format(
            self.__class__.__name__,
            self.service_name,
            self.resource_name,
            self.api_version
        )

    # Kinda ugly (method within a class definition, but not static/classmethod)
    # but depends on internal state. Grump.
    def requires_loaded(func):
        """
        A decorator to ensure the resource data is loaded.
        """
        def _wrapper(self, *args, **kwargs):
            # If we don't have data, go load it.
            if self._loaded_data is None:
                self._loaded_data = self.loader.load(self.service_name)

            return func(self, *args, **kwargs)

        return _wrapper

    @property
    @requires_loaded
    def service_data(self):
        """
        Returns all introspected service data. This will include things like
        other resources/collections that are part of the service. Typically,
        using ``.resource_data`` is much more useful/relevant.

        If the data has been previously accessed, a memoized version of the
        data is returned.

        :returns: A dict of introspected service data
        :rtype: dict
        """
        return self._loaded_data

    @property
    @requires_loaded
    def resource_data(self):
        """
        Returns all introspected resource data.

        If the data has been previously accessed, a memoized version of the
        data is returned.

        :returns: A dict of introspected resource data
        :rtype: dict
        """
        return self._loaded_data['resources'][self.resource_name]

    @property
    @requires_loaded
    def api_version(self):
        """
        Returns the API version introspected from the resource data.
        This is useful in preventing mismatching API versions between the
        client code & service.

        If the data has been previously accessed, a memoized version of the
        API version is returned.

        :returns: The service's version
        :rtype: string
        """
        self._api_version = self._loaded_data.get('api_version', '')
        return self._api_version

    @property
    @requires_loaded
    def identifiers(self):
        """
        Returns the identifiers.

        If the data has been previously accessed, a memoized version of the
        variable name is returned.

        :returns: The identifiers
        :rtype: list
        """
        return self.resource_data['identifiers']

    @requires_loaded
    def result_key_for(self, op_name):
        """
        Checks for the presence of a ``result_key``, which defines what data
        should make up an instance.

        Returns ``None`` if there is no ``result_key``.

        :param op_name: The operation name to look for the ``result_key`` in.
        :type op_name: string

        :returns: The expected key to look for data within
        :rtype: string or None
        """
        ops = self.resource_data.get('operations', {})
        op = ops.get(op_name, {})
        key = op.get('result_key', None)
        return key

    @property
    @requires_loaded
    def relations(self):
        """
        Returns the relation data read from the resource data.

        Example data::

            {
                'name_on_the_instance_here': {
                    'class_type': 'resource',
                    'class': 'NameOfResource',
                    'required': True,
                    'rel_type': '1-M'
                }
            }

        :returns: The relation data, if any is present
        :rtype: dict
        """
        return self.resource_data.get('relations', {})


class Resource(object):
    """
    A common base class for all the ``Resource`` objects.
    """
    def __init__(self, connection=None, **kwargs):
        """
        Creates a new ``Resource`` instance.

        :param connection: (Optional) Specifies what connection to use.
            By default, this is a matching ``Connection`` subclass provided
            by the ``session`` (i.e. within S3, ``BucketResource`` would get a
            ``S3Connection`` from the session).
        :type connection: <class kotocore.connection.Connection> **SUBCLASS**

        :param **kwargs: (Optional) Instance data to be specified on the
            instance itself.
        :type **kwargs: dict
        """
        # Tracks the *built* relations (actual instances).
        self._relations = {}
        # Tracks the scalar data on the resource.
        self._data = {}
        self._connection = connection

        for key, value in kwargs.items():
            self._data[key] = value

        if self._connection is None:
            self._connection = self._details.session.connect_to(
                self._details.service_name
            )

        # Now that we have a connection, we can update docstrings.
        self._update_docstrings()

    def __str__(self):
        return "{0}: {1} in {2}".format(
            self.__class__.__name__,
            self._details.service_name,
            self._connection.region_name
        )

    def __getattr__(self, name):
        """
        Attempts to return either the related object or instance data for a
        given name if available.

        :param name: The instance data's name
        :type name: string
        """
        # Check to see if the thing being requested is a known relation.
        if name in self._details.relations:
            # Check if we already have built version.
            if not name in self._relations:
                # There's not a previously built object.
                # Lazily build it & assign it here.
                self._relations[name] = self.build_relation(name)

            return self._relations[name]

        if name in self._data:
            return self._data[name]

        raise AttributeError("No such attribute '{0}'".format(name))

    def _update_docstrings(self):
        """
        Runs through the operation methods & updates their docstrings if
        necessary.

        If the method has the default placeholder docstring, this will replace
        it with the docstring from the underlying connection.
        """
        ops = self._details.resource_data['operations']

        for method_name in ops.keys():
            meth = getattr(self.__class__, method_name, None)

            if not meth:
                continue

            if meth.__doc__ != DEFAULT_DOCSTRING:
                # It already has a custom docstring. Leave it alone.
                continue

            # Needs updating. So there's at least *something* vaguely useful
            # there, use the docstring from the underlying ``Connection``
            # method.
            # FIXME: We need to figure out a way to make this more useful, if
            #        possible.
            api_name = ops[method_name]['api_name']
            conn_meth = getattr(self._connection, to_snake_case(api_name))

            # We need to do detection here, because Py2 treats ``.__doc__``
            # as a special read-only attribute. :/
            if six.PY3:
                meth.__doc__ = conn_meth.__doc__
            else:
                meth.__func__.__doc__ = conn_meth.__doc__

    def get_identifiers(self):
        """
        Returns the identifier(s) (if present) from the instance data.

        The identifier name(s) is/are determined from the ``ResourceDetails``
        instance hanging off the class itself.

        :returns: All the identifier information
        :rtype: dict
        """
        data = {}

        for id_info in self._details.identifiers:
            var_name = id_info['var_name']
            data[var_name] = self._data.get(var_name)

        return data

    def set_identifiers(self, data):
        """
        Sets the identifier(s) within the instance data.

        The identifier name(s) is/are determined from the ``ResourceDetails``
        instance hanging off the class itself.

        :param data: The value(s) to be set.
        :param data: dict
        """
        for id_info in self._details.identifiers:
            var_name = id_info['var_name']
            self._data[var_name] = data.get(var_name)

        # FIXME: This needs to likely kick off invalidating/rebuilding
        #        relations.
        #        For now, just remove them all. This is potentially inefficient
        #        but is nicely lazy if we don't need them & prevents stale data
        #        for the moment.
        self._relations = {}

    def build_relation(self, name, klass=None):
        """
        Constructs a related ``Resource`` or ``Collection``.

        This allows for construction of classes with information prepopulated
        from what the current instance has. This enables syntax like::

            bucket = Bucket(bucket='some-bucket-name')

            for obj in bucket.objects.each():
                print(obj.key)

        :param name: The name of the relation from the ResourceJSON
        :type name: string

        :param klass: (Optional) An overridable class to construct. Typically
            only useful if you need a custom subclass used in place of what
            kotocore provides.
        :type klass: class

        :returns: An instantiated related object
        """
        try:
            rel_data = self._details.relations[name]
        except KeyError:
            msg = "No such relation named '{0}'.".format(name)
            raise NoRelation(msg)

        if klass is None:
            # This is the typical case, where we're not explicitly given a
            # class to build with. Hit the session & look up what we should
            # be loading.
            if rel_data['class_type'] == 'collection':
                klass = self._details.session.get_collection(
                    self._details.service_name,
                    rel_data['class']
                )
            elif rel_data['class_type'] == 'resource':
                klass = self._details.session.get_resource(
                    self._details.service_name,
                    rel_data['class']
                )
            else:
                msg = "Unknown class '{0}' for '{1}'.".format(
                    rel_data['class_type'],
                    name
                )
                raise NoRelation(msg)

        # Instantiate & return it.
        kwargs = {}
        # Just populating identifiers is enough for the 1-M case.
        kwargs.update(self.get_identifiers())

        if rel_data.get('rel_type', '1-M') == '1-1':
            # FIXME: If it's not a collection, we might have some instance data
            #        (i.e. ``bucket``) in ``self._data`` to populate as well.
            #        This seems like a can of worms, so ignore for the moment.
            pass

        return klass(connection=self._connection, **kwargs)

    def full_update_params(self, conn_method_name, params):
        """
        When a API method on the resource is called, this goes through the
        params & run a series of hooks to allow for updating those parameters.

        Typically, this method is **NOT** call by the user. However, the user
        may wish to define other methods (i.e. ``update_params`` to work with
        multiple parameters at once or ``update_params_METHOD_NAME`` to
        manipulate a single parameter) on their class, which this method
        will call.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param params: A dictionary of all the key/value pairs passed to the
            method. This dictionary is transformed by this call into the final
            params to be passed to the underlying connection.
        :type params: dict
        """
        # We'll check for custom methods to do addition, specific work.
        custom_method_name = 'update_params_{0}'.format(conn_method_name)
        custom_method = getattr(self, custom_method_name, None)

        if custom_method:
            # Let the specific method further process the data.
            params = custom_method(params)

        # Now that all the method-specific data is there, apply any further
        # service-wide changes here.
        params = self.update_params(conn_method_name, params)
        return params

    def update_params(self, conn_method_name, params):
        """
        A hook to allow manipulation of multiple parameters at once.

        By default, this just ensures the identifier data in in the parameters,
        so that the user doesn't have to provide it.

        You can override/extend this method (typically on your subclass)
        to do additional checks, pre-populate values or remove unwanted data.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param params: A dictionary of all the key/value pairs passed to the
            method. This dictionary is transformed by this call into the final
            params to be passed to the underlying connection.
        :type params: dict
        """
        # By default, this just sets the identifier info.
        # We use ``var_name`` instead of ``api_name``. Because botocore.
        params.update(self.get_identifiers())
        return params

    def full_post_process(self, conn_method_name, result):
        """
        When a response from an API method call is received, this goes through
        the returned data & run a series of hooks to allow for handling that
        data.

        Typically, this method is **NOT** call by the user. However, the user
        may wish to define other methods (i.e. ``post_process`` to work with
        all the data at once or ``post_process_METHOD_NAME`` to
        handle a single piece of data) on their class, which this method
        will call.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param result: A dictionary of all the key/value pairs passed back
            from the API (server-side). This dictionary is transformed by this
            call into the final data to be passed back to the user.
        :type result: dict
        """
        result = self.post_process(conn_method_name, result)

        # We'll check for custom methods to do addition, specific work.
        custom_method_name = 'post_process_{0}'.format(conn_method_name)
        custom_method = getattr(self, custom_method_name, None)

        if custom_method:
            # Let the specific method further process the data.
            result = custom_method(result)

        return result

    def post_process(self, conn_method_name, result):
        """
        A hook to allow manipulation of the entire returned data at once.

        By default, this does nothing, just passing through the ``result``.

        You can override/extend this method (typically on your subclass)
        to do additional checks, alter the result or remove unwanted data.

        :param conn_method_name: The name of the underlying connection method
            about to be called. Typically, this is a "snake_cased" variant of
            the API name (i.e. ``update_bucket`` in place of ``UpdateBucket``).
        :type conn_method_name: string

        :param result: A dictionary of all the key/value pairs passed back
            from the API (server-side). This dictionary is transformed by this
            call into the final data to be passed back to the user.
        :type result: dict
        """
        # Mostly a hook for post-processing as needed.
        return result

    def post_process_get(self, result):
        """
        Given an object with identifiers, fetches the data for that object
        from the service.

        This alters the data on the object itself & simply passes through what
        was received.

        :param result: The response data
        :type result: dict

        :returns: The unmodified response data
        """
        if not hasattr(result, 'items'):
            # If it's not a dict, give up & just return whatever you get.
            return result

        # We need to possibly drill into the response & get out the data here.
        # Check for a result key.
        result_key = self._details.result_key_for('get')

        if not result_key:
            # There's no result_key. Just use the top-level data.
            data = result
        else:
            data = result[result_key]

        for key, value in data.items():
            self._data[to_snake_case(key)] = value

        return result


class ResourceFactory(object):
    """
    Generates the underlying ``Resource`` classes based off the ``ResourceJSON``
    included in the SDK.

    Usage::

        >>> rf = ResourceFactory()
        >>> Bucket = rf.construct_for('s3', 'Bucket')

    """
    loader_class = ResourceJSONLoader

    def __init__(self, session=None, loader=None,
                 base_resource_class=Resource,
                 details_class=ResourceDetails):
        """
        Creates a new ``ResourceFactory`` instance.

        :param session: The ``Session`` the factory should use.
        :type session: <class kotocore.session.Session> instance

        :param loader: (Optional) An instance of a ``ResourceJSONLoader`` class.
            This can be swapped with a different instance or with a completely
            different class with the same interface.
            By default, this is ``kotocore.loader.default_loader``.
        :type loader: <class kotocore.loader.ResourceJSONLoader> instance

        :param base_resource_class: (Optional) The base class to use when creating
            the resource. By default, this is ``Resource``, but should
            you need to globally change the behavior of all resources,
            you'd simply specify this to provide your own class.
        :type base_resource_class: <class kotocore.resources.Resource>

        :param details_class: (Optional) The metadata class used to store things
            like service name & data. By default, this is ``ResourceDetails``,
            but should you need to globally change the behavior (perhaps
            modifying how the resource data is returned), you simply provide
            your own class here.
        :type details_class: <class kotocore.resources.ResourceDetails>
        """
        self.session = session
        self.loader = loader
        self.base_resource_class = base_resource_class
        self.details_class = details_class

        if self.session is None:
            # Fallback to the default.
            import kotocore
            self.session = kotocore.session

        if self.loader is None:
            import kotocore.loader
            self.loader = kotocore.loader.default_loader

    def __str__(self):
        return self.__class__.__name__

    def construct_for(self, service_name, resource_name, base_class=None):
        """
        Builds a new, specialized ``Resource`` subclass as part of a given
        service.

        This will load the ``ResourceJSON``, determine the correct
        mappings/methods & constructs a brand new class with those methods on
        it.

        :param service_name: The name of the service to construct a resource
            for. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param resource_name: The name of the ``Resource``. Ex.
            ``Queue``, ``Notification``, ``Table``, etc.
        :type resource_name: string

        :returns: A new resource class for that service
        """
        details = self.details_class(
            self.session,
            service_name,
            resource_name,
            loader=self.loader
        )

        attrs = {
            '_details': details,
        }

        # Determine what we should call it.
        klass_name = self._build_class_name(resource_name)

        # Construct what the class ought to have on it.
        attrs.update(self._build_methods(details))

        if base_class is None:
            base_class = self.base_resource_class

        # Create the class.
        return type(
            klass_name,
            (base_class,),
            attrs
        )

    def _build_class_name(self, resource_name):
        return resource_name

    def _build_methods(self, details):
        attrs = {}
        ops = details.resource_data.get('operations', {}).items()

        for method_name, op_data in ops:
            attrs[method_name] = self._create_operation_method(
                method_name,
                op_data
            )

        return attrs

    def _create_operation_method(factory_self, method_name, op_data):
        # Determine the correct name for the method.
        # This is because the method names will be standardized across
        # resources, so we'll have to lean on the ``api_name`` to figure out
        # what the correct underlying method name on the ``Connection`` should
        # be.
        # Map -> map -> unmap -> remap -> map :/
        conn_method_name = to_snake_case(op_data['api_name'])

        if not six.PY3:
            method_name = str(method_name)

        def _new_method(self, **kwargs):
            params = self.full_update_params(method_name, kwargs)
            method = getattr(self._connection, conn_method_name, None)

            if not method:
                msg = "Introspected method named '{0}' not available on " + \
                      "the connection."
                raise NoSuchMethod(msg.format(conn_method_name))

            result = method(**params)
            return self.full_post_process(method_name, result)

        _new_method.__name__ = method_name
        _new_method.__doc__ = DEFAULT_DOCSTRING
        return _new_method



class BotoException(Exception):
    """
    A base exception class for all Boto-related exceptions.
    """
    pass


class ServerError(BotoException):
    """
    Thrown when an error is received within a response.
    """
    fmt = "[{0}]: {1}"

    def __init__(self, code='GeneralError', message='No message',
                 full_response=None, **kwargs):
        self.code = code
        self.message = message
        self.full_response = full_response

        if self.full_response is None:
            self.full_response = {}

        msg = self.fmt.format(
            self.code,
            self.message
        )
        super(ServerError, self).__init__(msg)


class IncorrectImportPath(BotoException):
    pass


class NoSuchMethod(BotoException):
    pass


class NotCached(BotoException):
    pass


class ResourceError(BotoException):
    pass


class NoResourceJSONFound(ResourceError):
    pass


class APIVersionMismatchError(BotoException):
    pass


class NoRelation(ResourceError):
    pass


class ValidationError(BotoException):
    pass


class MD5ValidationError(ValidationError):
    pass



#!/usr/bin/env python

"""
distutils/setuptools install script.
"""

import os
import sys
import kotocore

try:
    from setuptools import setup
    setup
except ImportError:
    from distutils.core import setup


packages = [
    'kotocore',
    'kotocore.utils',
]

requires = [
    'botocore>=0.24.0',
    'six>=1.4.0',
    'jmespath>=0.1.0',
    'python-dateutil>=2.1',
]

setup(
    name='kotocore',
    version=kotocore.get_version(),
    description='Utility for botocore.',
    long_description=open('README.rst').read(),
    author='Henry Huang',
    author_email='henry.s.huang@gmail.com',
    url='https://github.com/henrysher/kotocore',
    scripts=[],
    packages=packages,
    package_data={
        'kotocore': [
            'data/aws/resources/*.json',
        ]
    },
    include_package_data=True,
    install_requires=requires,
    license=open("LICENSE").read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
)



import botocore.session

from kotocore.cache import ServiceCache
from kotocore.utils.constants import USER_AGENT_NAME, USER_AGENT_VERSION
from kotocore.exceptions import NotCached


class Session(object):
    """
    Stores all the state for a given ``kotocore`` session.

    Can dynamically create all the various ``Connection`` classes.

    Usage::

        >>> from kotocore.session import Session
        >>> session = Session()
        >>> sqs_conn = session.connect_to('sqs', region_name='us-west-2')

    """
    cache_class = ServiceCache

    def __init__(self, session=None, connection_factory=None,
                 resource_factory=None, collection_factory=None):
        """
        Creates a ``Session`` instance.

        :param session: (Optional) Custom instantiated ``botocore`` instance.
            Useful if you have specific needs. If not present, a default
            ``Session`` will be created.
        :type session: <botocore.session.Session> instance

        :param connection_factory: (Optional) Specifies a custom
            ``ConnectionFactory`` to be used. Useful if you need to change how
            ``Connection`` objects are constructed by the session.
        :type connection_factory: <kotocore.connection.ConnectionFactory>
            instance

        :param resource_factory: (Optional) Specifies a custom
            ``ResourceFactory`` to be used. Useful if you need to change how
            ``Resource`` objects are constructed by the session.
        :type resource_factory: <kotocore.resources.ResourceFactory>
            instance

        :param collection_factory: (Optional) Specifies a custom
            ``CollectionFactory`` to be used. Useful if you need to change how
            ``Collection`` objects are constructed by the session.
        :type collection_factory: <kotocore.collections.CollectionFactory>
            instance
        """
        super(Session, self).__init__()
        self.core_session = session
        self.connection_factory = connection_factory
        self.resource_factory = resource_factory
        self.collection_factory = collection_factory

        self.cache = self.cache_class()

        if not self.core_session:
            self.core_session = botocore.session.get_session()

        self.core_session.user_agent_name = USER_AGENT_NAME
        self.core_session.user_agent_version = USER_AGENT_VERSION

        if not self.connection_factory:
            from kotocore.connection import ConnectionFactory
            self.connection_factory = ConnectionFactory(session=self)

        if not self.resource_factory:
            from kotocore.resources import ResourceFactory
            self.resource_factory = ResourceFactory(session=self)

        if not self.collection_factory:
            from kotocore.collections import CollectionFactory
            self.collection_factory = CollectionFactory(session=self)

    def get_connection(self, service_name):
        """
        Returns a ``Connection`` **class** for a given service.

        :param service_name: A string that specifies the name of the desired
            service. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :rtype: <kotocore.connection.Connection subclass>
        """
        try:
            return self.cache.get_connection(service_name)
        except NotCached:
            pass

        # We didn't find it. Construct it.
        new_class = self.connection_factory.construct_for(service_name)
        self.cache.set_connection(service_name, new_class)
        return new_class

    def get_resource(self, service_name, resource_name, base_class=None):
        """
        Returns a ``Resource`` **class** for a given service.

        :param service_name: A string that specifies the name of the desired
            service. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param resource_name: A string that specifies the name of the desired
            class. Ex. ``Queue``, ``Notification``, ``Table``, etc.
        :type resource_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
        :type base_class: class

        :rtype: <kotocore.resources.Resource subclass>
        """
        try:
            return self.cache.get_resource(
                service_name,
                resource_name,
                base_class=base_class
            )
        except NotCached:
            pass

        # We didn't find it. Construct it.
        new_class = self.resource_factory.construct_for(
            service_name,
            resource_name,
            base_class=base_class
        )
        self.cache.set_resource(service_name, resource_name, new_class)
        return new_class

    def get_collection(self, service_name, collection_name, base_class=None):
        """
        Returns a ``Collection`` **class** for a given service.

        :param service_name: A string that specifies the name of the desired
            service. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: A string that specifies the name of the desired
            class. Ex. ``QueueCollection``, ``NotificationCollection``,
            ``TableCollection``, etc.
        :type collection_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
        :type base_class: class

        :rtype: <kotocore.collections.Collection subclass>
        """
        try:
            return self.cache.get_collection(
                service_name,
                collection_name,
                base_class=base_class
            )
        except NotCached:
            pass

        # We didn't find it. Construct it.
        new_class = self.collection_factory.construct_for(
            service_name,
            collection_name,
            base_class=base_class
        )
        self.cache.set_collection(service_name, collection_name, new_class)
        return new_class

    def connect_to(self, service_name, **kwargs):
        """
        Shortcut method to make instantiating the ``Connection`` classes
        easier.

        Forwards ``**kwargs`` like region, keys, etc. on to the constructor.

        :param service_name: A string that specifies the name of the desired
            service. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :rtype: <kotocore.connection.Connection> instance
        """
        service_class = self.get_connection(service_name)
        return service_class.connect_to(**kwargs)

    def get_core_service(self, service_name):
        """
        Returns a ``botocore.service.Service``.

        Mostly an abstraction for the ``*Connection`` objects to get what
        they need for introspection.

        :param service_name: A string that specifies the name of the desired
            service. Ex. ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :rtype: <botocore.service.Service subclass>
        """
        return self.core_session.get_service(service_name)



"""
Some commonly used things that aren't bundled with (all versions of) Python.

We'll lean on botocore for them, to limit the amount of duplication.
"""
from botocore.compat import json
from botocore.compat import OrderedDict
from botocore.compat import six



import logging


__author__ = 'Amazon Web Services'
__version__ = (0, 1, 2, 'alpha')


def get_version(full=False):
    """
    Returns a string-ified version number.

    Optionally accepts a ``full`` parameter, which if ``True``, will include
    any pre-release information. (Default: ``False``)
    """
    version = '.'.join([str(bit) for bit in __version__[:3]])

    if full:
        version = '-'.join([version] + list(__version__[3:]))

    return version


# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


log = logging.getLogger('kotocore')
log.addHandler(NullHandler())
# End logging setup.



from kotocore.exceptions import NotCached


class ServiceCache(object):
    """
    A centralized registry of classes that have already been built.

    Present to both prevent too much factory churn as well as to give the
    resource layer something to refer to in relations.

    Usage::

        >>> sc = ServiceCache()
        >>> len(sc)
        0
        >>> sc.set_connection('s3', S3Connection)
        >>> sc.set_resource('s3', 'Bucket', Bucket)
        >>> sc.set_collection('s3', 'BucketCollection', BucketCollection)
        # We only count services.
        >>> len(sc)
        1
        >>> 's3' in sc
        True
        # Later...
        >>> conn_class = sc.get_connection('s3')
        >>> res_class = sc.get_resource('s3', 'Bucket')
        >>> sc.del_resource('s3', 'Bucket')

    """
    # TODO: We may want to add LRU/expiration behavior in the future, to
    #       prevent the cache from taking up too much space.
    #       Unlikely, but potential.
    def __init__(self):
        self.services = {}

    def __str__(self):
        return 'ServiceCache: {0}'.format(
            ', '.join(sorted(self.services.keys()))
        )

    def __len__(self):
        return len(self.services)

    def __contains__(self, service_name):
        return service_name in self.services

    def get_connection(self, service_name):
        """
        Retrieves a connection class from the cache, if available.

        :param service_name: The service a given ``Connection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :returns: A <kotocore.connection.Connection> subclass
        """
        service = self.services.get(service_name, {})
        connection_class = service.get('connection', None)

        if not connection_class:
            msg = "Connection for '{0}' is not present in the cache."
            raise NotCached(msg.format(
                service_name
            ))

        return connection_class

    def set_connection(self, service_name, to_cache):
        """
        Sets a connection class within the cache.

        :param service_name: The service a given ``Connection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param to_cache: The class to be cached for the service.
        :type to_cache: class
        """
        self.services.setdefault(service_name, {})
        self.services[service_name]['connection'] = to_cache

    def del_connection(self, service_name):
        """
        Deletes a connection for a given service.

        Fails silently if no connection is found in the cache.

        :param service_name: The service a given ``Connection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string
        """
        # Unlike ``get_connection``, this should be fire & forget.
        # We don't really care, as long as it's not in the cache any longer.
        try:
            del self.services[service_name]['connection']
        except KeyError:
            pass

    def build_classpath(self, klass=None):
        if not klass:
            classpath = 'default'
        else:
            classpath = "{0}.{1}".format(
                klass.__module__,
                klass.__name__
            )

        return classpath

    def get_resource(self, service_name, resource_name, base_class=None):
        """
        Retrieves a resource class from the cache, if available.

        :param service_name: The service a given ``Resource`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param resource_name: The name of the ``Resource``. Ex.
            ``Queue``, ``Notification``, ``Table``, etc.
        :type resource_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
            Default is ``default``.
        :type base_class: class

        :returns: A <kotocore.resources.Resource> subclass
        """
        classpath = self.build_classpath(base_class)
        service = self.services.get(service_name, {})
        resources = service.get('resources', {})
        resource_options = resources.get(resource_name, {})
        resource_class = resource_options.get(classpath, None)

        if not resource_class:
            msg = "Resource '{0}' for {1} is not present in the cache."
            raise NotCached(msg.format(
                resource_name,
                service_name
            ))

        return resource_class

    def set_resource(self, service_name, resource_name, to_cache):
        """
        Sets the resource class within the cache.

        :param service_name: The service a given ``Resource`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param resource_name: The name of the ``Resource``. Ex.
            ``Queue``, ``Notification``, ``Table``, etc.
        :type resource_name: string

        :param to_cache: The class to be cached for the service.
        :type to_cache: class
        """
        self.services.setdefault(service_name, {})
        self.services[service_name].setdefault('resources', {})
        self.services[service_name]['resources'].setdefault(resource_name, {})
        options = self.services[service_name]['resources'][resource_name]
        classpath = self.build_classpath(to_cache.__bases__[0])

        if classpath == 'kotocore.resources.Resource':
            classpath = 'default'

        options[classpath] = to_cache

    def del_resource(self, service_name, resource_name, base_class=None):
        """
        Deletes a resource class for a given service.

        Fails silently if no connection is found in the cache.

        :param service_name: The service a given ``Resource`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
            Default is ``default``.
        :type base_class: class
        """
         # Unlike ``get_resource``, this should be fire & forget.
        # We don't really care, as long as it's not in the cache any longer.
        try:
            classpath = self.build_classpath(base_class)
            opts = self.services[service_name]['resources'][resource_name]
            del opts[classpath]
        except KeyError:
            pass

    def get_collection(self, service_name, collection_name, base_class=None):
        """
        Retrieves a collection class from the cache, if available.

        :param service_name: The service a given ``Collection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: The name of the ``Collection``. Ex.
            ``QueueCollection``, ``NotificationCollection``,
            ``TableCollection``, etc.
        :type collection_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
            Default is ``default``.
        :type base_class: class

        :returns: A <kotocore.collections.Collection> subclass
        """
        classpath = self.build_classpath(base_class)
        service = self.services.get(service_name, {})
        collections = service.get('collections', {})
        collection_options = collections.get(collection_name, {})
        collection_class = collection_options.get(classpath, None)

        if not collection_class:
            msg = "Collection '{0}' for {1} is not present in the cache."
            raise NotCached(msg.format(
                collection_name,
                service_name
            ))

        return collection_class

    def set_collection(self, service_name, collection_name, to_cache):
        """
        Sets a collection class within the cache.

        :param service_name: The service a given ``Collection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: The name of the ``Collection``. Ex.
            ``QueueCollection``, ``NotificationCollection``,
            ``TableCollection``, etc.
        :type collection_name: string

        :param to_cache: The class to be cached for the service.
        :type to_cache: class
        """
        self.services.setdefault(service_name, {})
        self.services[service_name].setdefault('collections', {})
        self.services[service_name]['collections'].setdefault(collection_name, {})
        options = self.services[service_name]['collections'][collection_name]
        classpath = self.build_classpath(to_cache.__bases__[0])

        if classpath == 'kotocore.collections.Collection':
            classpath = 'default'

        options[classpath] = to_cache

    def del_collection(self, service_name, collection_name, base_class=None):
        """
        Deletes a collection for a given service.

        Fails silently if no collection is found in the cache.

        :param service_name: The service a given ``Collection`` talks to. Ex.
            ``sqs``, ``sns``, ``dynamodb``, etc.
        :type service_name: string

        :param collection_name: The name of the ``Collection``. Ex.
            ``QueueCollection``, ``NotificationCollection``,
            ``TableCollection``, etc.
        :type collection_name: string

        :param base_class: (Optional) The base class of the object. Prevents
            "magically" loading the wrong class (one with a different base).
            Default is ``default``.
        :type base_class: class
        """
         # Unlike ``get_collection``, this should be fire & forget.
        # We don't really care, as long as it's not in the cache any longer.
        try:
            classpath = self.build_classpath(base_class)
            opts = self.services[service_name]['collections'][collection_name]
            del opts[classpath]
        except KeyError:
            pass
